from datetime import timedelta
import re
from shlex import split


def parse_timespec(timespec: str) -> timedelta:
    """
    returns a negative timedelta corresponding to the given timespec
    :param timespec: a space-delimited string of time info. see examples
    :return:
    """
    days = 0
    seconds = 0
    for e in timespec.split():
        if e[-1:] == 'd':
            days -= int(e[:-1])
        if e[-1:] == 'M':
            days -= int(e[:-1]) * 30
        if e[-1:] == 'y':
            days -= int(e[:-1]) * 365
        if e[-1:] == 'h':
            seconds -= int(e[:-1]) * 3600
        if e[-1:] == 'm':
            seconds -= int(e[:-1]) * 60
        if e[-1:] == 's':
            seconds -= int(e[:-1])
    return timedelta(days=days, seconds=seconds)


class InvalidConfigurationError(RuntimeError):
    pass


class Config(object):
    """
    Parses a config file in a special format
    """

    class Context(object):
        def __init__(self):
            self.glob = True
            self.account = None
            self.mailbox = None

    def __init__(self):
        self.config = dict()

    def get_accounts(self):
        return [k for k in self.config.keys()
                if type(self.config[k]) == dict]

    def get_mailboxes(self, account):
        return [k for k in self.config[account].keys()
                if type(self.config[account][k]) == dict]

    def parse(self, path: str):
        ctxt = Config.Context()
        with open(path) as f:
            for line in f.readlines():
                self.parse_line(line, ctxt)

    def parse_line(self, line: str, context: Context):
        """
        >>> c = Config()
        >>> c.parse_line("# terte  ", Config.Context())
        >>> c.parse_line('    # dfgj', Config.Context())
        >>> ctx = Config.Context()
        >>> ctx.glob
        True
        >>> ctx.account
        >>> ctx.mailbox
        >>> c.parse_line("testmb {", ctx)
        >>> ctx.glob
        False
        >>> ctx.account
        'testmb'
        >>> ctx.mailbox
        >>> c.parse_line("refresh = 1", ctx)
        >>> c.config['testmb']["refresh"]
        '1'
        >>> c.parse_line(' }  ', ctx)
        >>> ctx.account
        >>> ctx.glob
        True


        :param line:
        :param context:
        :return:
        """
        if line.strip().startswith('#'):
            # Then it is a comment, we ignore it
            return
        match = re.match('^ *(.*) \{$', line)
        if match:
            if context.glob:
                if context.account is not None or context.mailbox is not None:
                    raise InvalidConfigurationError('error near %s' % line)
                context.glob = False
                context.account = match.group(1)
                self.config[context.account] = {}
                return
            if context.account:
                if context.glob:
                    raise InvalidConfigurationError('error near %s' % line)
                context.mailbox = match.group(1)
                self.config[context.account][context.mailbox] = {}
                self.config[context.account][context.mailbox]['rules'] = []
                return
        match = re.match('^ *(?P<key>.*) = (?P<value>.*)$', line)
        if match:
            key, value = match.group('key'), match.group('value')
            if context.glob:
                self.config[key] = value
            elif context.account and not context.mailbox:
                self.config[context.account][key] = value
            else:
                self.config[context.account][context.mailbox][key] = value
        if line.strip().startswith('if '):
            self.config[context.account][context.mailbox]['rules'].append(
                Config.parse_rule(line))
        if line.strip().startswith('}'):
            if context.mailbox:
                context.mailbox = None
                return
            if context.account:
                context.account = None
                context.glob = True
                return
            return InvalidConfigurationError('got } in an global context')

    def get(self, key, default=None, mb=None, account=None):
        """
        >>> conf = '''a = 1
        ... c = 1
        ... ta1 {
        ...   a = 2
        ...   b = 1
        ...   tmb1 {
        ...     b = 2
        ...     c = 2
        ...     d = 1
        ...   }
        ... }'''
        >>> c = Config()
        >>> ctx = Config.Context()
        >>> for line in conf.splitlines():
        ...     c.parse_line(line, ctx)
        >>> c.get('a', default='test', mb='tmb1', account='ta1')
        '2'
        >>> c.get('b', default='test', mb='tmb1', account='ta1')
        '2'
        >>> c.get('c', default='test', mb='tmb1', account='ta1')
        '2'
        >>> c.get('d', default='test', mb='tmb1', account='ta1')
        '1'
        >>> c.get('a', default='test', account='ta1')
        '2'
        >>> c.get('b', default='test', account='ta1')
        '1'
        >>> c.get('c', default='test', account='ta1')
        '1'
        >>> c.get('d', default='test', account='ta1')
        'test'
        >>> c.get('a', default='test', account='ta1')
        '2'
        >>> c.get('b', default='test', account='ta1')
        '1'
        >>> c.get('c', default='test', account='ta1')
        '1'
        >>> c.get('d', default='test', account='ta1')
        'test'
        >>> c.get('a', default='test')
        '1'
        >>> c.get('b', default='test')
        'test'
        >>> c.get('c', default='test')
        '1'
        >>> c.get('d', default='test')
        'test'

        :param key: the key to fetch
        :param default: a default value to return if the key is not found
        :param mb: name of the mailbox
        :param account: name of the account
        :return: something
        """
        if account:
            if mb:
                if key in self.config[account][mb].keys():
                    return self.config[account][mb][key]
                if key in self.config[account].keys():
                    return self.config[account][key]
                if key in self.config.keys():
                    return self.config[key]
            if key in self.config[account].keys():
                return self.config[account][key]
            if key in self.config.keys():
                return self.config[key]
        if key in self.config.keys():
            return self.config[key]
        return default

    @classmethod
    def parse_rule(cls, line: str) -> dict:
        """
        parses a rule under the form of

            if is_from 'sample@test.net' and read then archive

        >>> from pprint import pprint
        >>> pprint(Config.parse_rule("read then archive"))
        {'actions': [['archive']], 'conditions': [['read']]}
        >>> pprint(Config.parse_rule("if is_from sample@test.net and read then archive"))
        {'actions': [['archive']],
         'conditions': [['is_from', 'sample@test.net'], ['read']]}
        >>> pprint(Config.parse_rule("if is_from sample@test.net and read then unstar and archive"))
        {'actions': [['unstar'], ['archive']],
         'conditions': [['is_from', 'sample@test.net'], ['read']]}
        >>> pprint(Config.parse_rule("if is_from 'Hallo leute' then star"))
        {'actions': [['star']], 'conditions': [['is_from', 'Hallo leute']]}

        :param line:
        :return: a dict representing the rule
        """
        if line.strip().startswith('if '):
            line = line.strip()[3:]
        conditions, actions = line.split('then')
        conditions = conditions.split('and')
        actions = actions.split('and')
        rule = dict()
        rule['conditions'] = []
        rule['actions'] = []
        for element in conditions:
            rule['conditions'].append(split(element.strip('"')))
        for element in actions:
            rule['actions'].append(split(element.strip('"')))
        return rule



from sys import argv
from os.path import expanduser, join
from argparse import ArgumentParser

import logging

from . import __version__
from .bureaucrate import init
from .utils import Config


def process_account(conf: Config, account: str):
    acc = init(join(conf.get("base_path"), account))
    for mailbox in conf.get_mailboxes(account):
        for message in acc[mailbox]:
            message.exec_rules(conf.get('rules', [], mailbox, account))


def main():
    parser = ArgumentParser()
    parser.add_argument('--version', help="returns the version and exists")
    parser.add_argument('-a', '--account', help="Restrict to an account")
    parser.add_argument('-c', '--config', default='~/.bureaucraterc',
                        help='specify an alternate configuration file')
    parser.add_argument('--debug', dest='loglevel', help='enable debug logging',
                        action='store_const', const=logging.DEBUG,
                        default=logging.WARNING)
    opts = vars(parser.parse_args(argv[1:]))
    conf = Config()
    conf.parse(expanduser(opts.get('config')))

    from .bureaucrate import logger
    logger.setLevel(opts.get('loglevel'))

    if opts.get('version', None):
        print("bureaucrate v{}".format(__version__))
        return

    if opts.get('account', None):
        process_account(conf, opts.get('account'))
    else:
        for account in conf.get_accounts():
            process_account(conf, account)

if '__main__' in __name__:
    main()



# -*- coding: utf-8 -*-

import re
from datetime import datetime
from email.header import decode_header
from functools import wraps
from mailbox import Maildir, MaildirMessage
from os import listdir
from os.path import join, expanduser
from subprocess import run, PIPE
from typing import List, Optional, Dict

from chardet import detect
from dateutil.parser import parse as dateparse
from logging import getLogger, basicConfig


from .utils import parse_timespec


logger = getLogger(__name__)
FORMAT = "%(filename)s:%(lineno)3s - %(funcName)15s - %(levelname)s: %(" \
         "message)s"
basicConfig(format=FORMAT)


class ConditionError(Exception):
    pass


class Mailbox(Maildir):
    def __iter__(self):
        for k, v in self.iteritems():
            v.mailbox = self
            v.key = k
            yield v


base_path = str
mailboxes = Dict[str, Mailbox]


def get_mailbox(mb_id: str) -> Mailbox:
    """
    gets or creates a Mailbox representation of the directory at
    base_path + mb_id
    :param mb_id:
    :return:
    """
    global mailboxes
    if mb_id not in mailboxes:
        mb = Mailbox(join(base_path, mb_id), factory=Message.message_factory)
        mailboxes[mb_id] = mb
    else:
        mb = mailboxes[mb_id]
    return mb


def init(mailbox_base: str, mailbox_names: List[str]=[]) -> Dict[str, Mailbox]:
    logger.info('Initializing mailboxes at %s', mailbox_base)
    global mailboxes, base_path
    base_path = expanduser(mailbox_base)
    if not mailbox_names:
        mailbox_names = [e for e in listdir(base_path)]
    mailboxes = {}
    for mailbox_name in mailbox_names:
        mailboxes[mailbox_name] = Mailbox(join(mailbox_base, mailbox_name),
                                          factory=Message.message_factory)
        logger.debug('mailbox found in %s: %s', mailbox_base, mailbox_name)
    return mailboxes


def condition(f):
    """
    Decorator for conditions
    """

    @wraps(f)
    def try_execute(*args, **kwargs):
        try:
            res, m = f(*args, **kwargs)
            m.conditions_results.append(res)
            return m
        except Exception as e:
            raise ConditionError(e)
    return try_execute


def action(f):
    """
    decorator for actions. Reinitialises the conditions_result array after action execution
    """

    @wraps(f)
    def decorate(message, *args, **kwargs):
        r = None
        if message.conditions_results.count(True) == \
            len(message.conditions_results):
            r = f(message, *args, **kwargs)
        return r or message

    return decorate


class Message(MaildirMessage):
    """
    Represents a mail message, and its associated conditions and actions
    """

    def __getitem__(self, name):
        "Let's convert headers to str"
        message = []
        if name not in self:
            logger.debug("name not found: %s", name)
            return ""
        for t in decode_header(super().__getitem__(name)):
            logger.debug("t: %s", t)
            if type(t[0]) is bytes:
                if t[1] == 'unknown-8bit':
                    logger.info('found unknown charset: %s', t)
                    try:
                        supposition = detect(t[0])
                        logger.debug('trying to recover as %s', supposition)
                        message.append(t[0].decode(supposition['encoding']))
                    except:
                        logger.error('Failed to recover "%s" as %s', t[0], supposition)
                        raise ValueError('Incorrect encoding for Message: %s' %
                                         super().__getitem__('subject'))
                else:
                    message.append(t[0].decode(t[1] or 'ASCII'))
            if type(t[0]) is str:
                message.append(t[0])
        return ''.join(message)

    @staticmethod
    def message_factory(message):
        m = Message(message)
        logger.debug('Handling mail from %s with subject "%s"', m['From'],
                     m['Subject'])
        logger.debug(m['Date'])
        try:
            d = dateparse(m['Date'])
        except ValueError:
            logger.warning('Mail from %s with subject "%s" Does not have a '
                           '"Date" header! Some features will fail because of '
                           'this!', m['From'], m['Subject'])
            d = datetime.now()
        m.set_date(d.timestamp())
        return m

    def __init__(self, *args, **kwargs):
        self.conditions_results = []
        self.mailbox = Optional[Mailbox]
        super().__init__(*args, **kwargs)

    def exec_rule(self, rule):
        for cond in rule['conditions']:
            logger.debug("eval cond: %s", cond)
            if len(cond) > 1:
                getattr(self, cond[0])(*cond[1:])
            else:
                getattr(self, cond[0])()
        for act in rule['actions']:
            logger.debug("eval action: %s", act)
            if len(act) > 1:
                getattr(self, act[0])(*act[1:])
            else:
                getattr(self, act[0])()

    def exec_rules(self, rules: List):
        for rule in rules:
            self.exec_rule(rule)
            self.conditions_results = []

    def get_list(self) -> str:
        s = str(self.get('list-id', ''))
        return s[s.find('<'):s.find('>')].strip('<').strip('>')

    @condition
    def negate(self):
        # type: () -> (bool, Message)
        """
        Changes the result of last condition to the opposite of what it was
        """
        self.conditions_results[-1] = not self.conditions_results[-1]
        return True, self

    @condition
    def is_from(self, target: str):
        """
        Returns True if target is in From: header

        >>> m = Message()
        >>> m['From'] = "Sample <example@test.org>"
        >>> _ = m.is_from('Sample')
        >>> m.conditions_results[-1]
        True
        >>> _ = m.is_from('example')
        >>> m.conditions_results[-1]
        True
        >>> _ = m.is_from('@test.org')
        >>> m.conditions_results[-1]
        True
        >>> _ = m.is_from('Paul')
        >>> m.conditions_results[-1]
        False
        """
        # type: (str) -> (bool, Message)
        return target in self['from'], self

    @condition
    def starred(self):
        """
        Returns True if message is flagged or 'starred'

        >>> m = Message()
        >>> _ = m.starred()
        >>> m.conditions_results[-1]
        False
        >>> m.set_flags('F')
        >>> _ = m.starred()
        >>> m.conditions_results[-1]
        True
        """
        # type: () -> (bool, Message)
        return 'F' in self.get_flags(), self

    @condition
    def read(self):
        """
        Returns True if message is seen or 'read'

        >>> m = Message()
        >>> _ = m.read()
        >>> m.conditions_results[-1]
        False
        >>> m.conditions_results = [] # reset conditional action exec
        >>> m.conditions_results
        []
        >>> _ = m.mark_as_read()
        >>> 'S' in m.get_flags()
        True
        >>> _ = m.read()
        >>> m.conditions_results[-1]
        True
        """
        # type: () -> (bool, Message)
        return 'S' in self.get_flags(), self

    @condition
    def subject_has(self, subject: str):
        """
        Returns True if target is in Subject: header

        >>> m = Message()
        >>> m['Subject'] = "Hi there!"
        >>> _ = m.subject_has('Hi')
        >>> m.conditions_results[-1]
        True
        >>> _ = m.subject_has('there')
        >>> m.conditions_results[-1]
        True
        >>> _ = m.subject_has('Hello')
        >>> m.conditions_results[-1]
        False
        """
        # type: (str) -> (bool, Message)
        return subject in self['subject'], self

    @condition
    def older_than(self, timespec: str):
        """
        Return True if message is older than 'timespec'.

        'timespec' is a simplistic expression of time span, in the form of '\d+[YMDhms]( (\d+[YMDhms]))*'
        """
        # type: (str) -> (bool, Message)
        now = datetime.now()
        d = now + parse_timespec(timespec)
        if datetime.fromtimestamp(self.get_date()) < d:
            return True, self
        return False, self

    @condition
    def is_list(self):
        """
        Returns True if the message is destined to a list, and if the corresponding List-Id: header is set
        """
        # type: () -> (bool, Message)
        return "List-Id" in self.keys(), self

    @condition
    def list_is(self, list: str):
        "Returns True if substring 'list' can be found in header List-Id"
        # type: (str) -> (bool, Message)
        if self.get_list():
            return list in self.get_list(), self
        return False, self

    @condition
    def has_replied(self):
        """
        Returns True if the message has been replied to
        :return:
        """
        # type: () -> (bool, Message)
        return 'R' in self.get_flags(), self

    @condition
    def is_spam(self):
        """
        :return: True if has a Spam header
        """
        if 'X-Spam' in self:
            if re.match(r'[Yy]es|[Tt]rue', self['X-Spam']):
                return True, self
        return False, self

    # actions
    @action
    def mark_as_read(self):
        """
        >>> m = Message()
        >>> 'S' in m.get_flags()
        False
        >>> _ = m.mark_as_read()
        >>> 'S' in m.get_flags()
        True

        :return:
        """
        # type: () -> Message
        self.add_flag('S')
        return self

    @action
    def star(self):
        # type: () -> Message
        self.add_flag('F')
        return self

    @action
    def delete(self) -> None:
        """
        Warning! Deletion is final and will break the execution chain!
        :return: None
        """
        self.add_flag('T')
        self.mailbox.remove(self.key)
        return None

    @action
    def move_to(self, box: str):
        # type: (Mailbox) -> Message
        box = get_mailbox(box)
        key = box.add(self)
        self.delete()
        return box.get(key)

    @action
    def copy(self, box: str):
        # type: (Mailbox) -> Message
        box = get_mailbox(box)
        key = box.add(self)
        return box.get(key)

    @action
    def archive(self, archive_format='Archives.%Y'):
        dt = datetime.fromtimestamp(self.get_date())
        return self.move_to(dt.strftime(archive_format))

    @action
    def forward(self, command, mto, mfrom=None):
        m = Message()
        m['From'] = mfrom or self['to']
        m['To'] = mto
        m['Subject'] = "Fwd: " + self['subject']
        m.set_payload(self.get_payload())
        run(command + " " + m['To'],
            stdout=PIPE, input=m.as_string().encode('utf-8'),
            shell=True, check=True)
        return self



import unittest
from bureaucrate.utils import parse_timespec
from datetime import timedelta


class TestUtils(unittest.TestCase):
    def test_timespec(self):
        d = timedelta(days=-6)
        self.assertEqual(parse_timespec('6d'), d)
        d = timedelta(days=-36, seconds=-630)
        self.assertEqual(parse_timespec('1M 6d 10m 30s'), d)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

with open('requirements.txt') as reqs_file:
    requirements = reqs_file.readlines()

with open('requirements_dev.txt') as devreqs_file:
    test_requirements = devreqs_file.readlines()

setup(
    name='bureaucrate',
    version='0.3.8',
    description="A maildir-based executer of rules, destined to sort and automate mail",
    long_description=readme + '\n\n' + history,
    author="Paul Ollivier",
    author_email='contact@paulollivier.fr',
    url='https://github.com/paulollivier/bureaucrate',
    packages=[
        'bureaucrate',
    ],
    package_dir={'bureaucrate':
                 'bureaucrate'},
    entry_points={
        'console_scripts': [
            'bureaucrate=bureaucrate.__main__:main'
        ]
    },
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    zip_safe=False,
    keywords='bureaucrate',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Environment :: Console',
        'Topic :: Communications :: Email :: Filters',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    test_suite='tests',
    tests_require=test_requirements
)
